/*
 * This file is part of IMGU.
 * 
 * @Copyright 2008 Université de Montréal, Laboratoire Vision3D
 *   Sébastien Roy (roys@iro.umontreal.ca)
 *   Vincent Chapdelaine-Couture (chapdelv@iro.umontreal.ca)
 *   Lucie Bélanger
 *   Jamil Draréni 
 *
 * IMGU is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * IMGU is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with IMGU.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "imgu.h"

int camAllocate(camera **cam,imgu *img)
{
    if ((*cam)==NULL)
    {
      (*cam)=(camera *)(malloc(sizeof(camera)));
      (*cam)->features=NULL;
      (*cam)->next=NULL;
    }

    (*cam)->img=img;

    mat4Identity((*cam)->M);
    mat4Identity((*cam)->Minv);
    mat4Identity((*cam)->R);
    mat4Identity((*cam)->T);
    mat3Identity((*cam)->K);
    mat3Identity((*cam)->Kinv);

    (*cam)->position[0]=0.0;
    (*cam)->position[1]=0.0;
    (*cam)->position[2]=0.0;
    (*cam)->orientation[0]=0.0;
    (*cam)->orientation[1]=0.0;
    (*cam)->orientation[2]=0.0;

    (*cam)->k1=(*cam)->k2=0;

    return 0;
}

int camAllocateMulti(camera **cam,imgu *ims)
{
  if (ims==NULL)
  {
    camFreeMulti(cam); //in case camAllocateMulti was called before with a longer ims list, we need to free the remaining structures
    return 0;
  }

  camAllocate(cam,ims);
  return  camAllocateMulti(&((*cam)->next),ims->next);
}

void camFree(camera **cam)
{
    if (cam!=NULL && (*cam)!=NULL)
    {
      matFree(&((*cam)->features));
      matFree(&((*cam)->features));
      free((*cam));
      (*cam)=NULL;
    }
}

void camFreeMulti(camera **cam)
{
    if (cam==NULL) return;

    camera *ctemp;

    while( *cam!=NULL ) {
  	  // free *cam, but save the next link before its too late...
	  ctemp=(*cam)->next;
	  camFree(cam);
	  *cam=ctemp;
    }
}

int camCopyParams(camera *dest,camera *src)
{
    if (dest==NULL) return -1;
    if (src==NULL) return -1;

    vectxCopy(src->M,dest->M,16);
    vectxCopy(src->K,dest->K,9);
    vectxCopy(src->Minv,dest->Minv,16);
    vectxCopy(src->Kinv,dest->Kinv,9);
    vectxCopy(src->orientation,dest->orientation,3);
    vectxCopy(src->position,dest->position,3);
    vectxCopy(src->T,dest->T,16);
    vectxCopy(src->R,dest->R,16);

    dest->k1=src->k1;
    dest->k2=src->k2;

    return 0;
}

void camDump(camera *cam)
{
    if (cam==NULL) return;
 
    printf("###################\n");
    printf("M:\n");
    mat4Print(cam->M);
    //mat4Print(cam->Minv);
    printf("K:\n");
    mat3Print(cam->K);
    //mat3Print(cam->Kinv);
    printf("Position:\n");
    vect3Print(cam->position);
    printf("Orientation:\n");
    vect3Print(cam->orientation);
    //mat4Print(cam->T);
    //mat4Print(cam->R);
    printf("Radial distortion:\n");
    printf("%f %f\n",cam->k1,cam->k2);
    printf("###################\n");
}

int camCount(camera *cam)
{
    int count;
    camera *curr;
    curr=cam;
    count=0;
    while(curr!=NULL)
    {
        curr=curr->next;
        count++;
    }

    return count;
}

/* Assumes position Tx,Ty,Tz of camera and its rotation Rx,Ry,Rz quaternion format are known*/
int camSetExternalParams(camera *cam,vector3 t,vector3 w)
{
    matrix4 T,Tinv,R,RT;
    if (cam==NULL) return -1;

    vectxCopy(w,cam->orientation,3);
    vectxCopy(t,cam->position,3);
	mat4Translation(t,T);
    mat4Inverse(T,Tinv);

	mat4RotationFromAxis(w,R);
    mat4Transpose(R,RT);
	mat4Multiply(RT,Tinv,cam->M);

    //P=TR
    mat4Copy(RT,cam->R);
    mat4Identity(cam->T);
    cam->T[3]=cam->M[3];
    cam->T[7]=cam->M[7];
    cam->T[11]=cam->M[11];

	if( mat4Inverse(cam->M,cam->Minv) ) { printf("Impossible d'inverser.\n");return(-1); }

    return 0;
}

int camSetExternalParamsFromMatrix(camera *cam,matrix4 Mext)
{
  matrix4 Tinv,R,RT;
  vector3 t,w;

  if (cam==NULL) return -1;

  mat4Copy(Mext,RT);
  RT[3]=0;
  RT[7]=0;
  RT[11]=0;
  mat4Transpose(RT,R);
  mat4Multiply(R,Mext,Tinv);

  mat4RotationToAxis(R,w);

  t[0]=-Tinv[3];
  t[1]=-Tinv[7];
  t[2]=-Tinv[11];

  camSetExternalParams(cam,t,w);

  return 0;
}

int camSetInternalParams(camera *cam,double fx,double fy,double skew,double ox,double oy)
{
    if (cam==NULL) return -1;

    cam->K[0]=fx;
    cam->K[1]=skew;
    cam->K[2]=ox;
    cam->K[3]=0.0;
    cam->K[4]=fy;
    cam->K[5]=oy;
    cam->K[6]=0.0;
    cam->K[7]=0.0;
    cam->K[8]=1.0;

	if( mat3Inverse(cam->K,cam->Kinv) ) { printf("Impossible d'inverser.\n");return(-1); }

    return 0;
}

int camSetInternalParamsFromViewAngle(camera *cam,double hangle)
{
    double angle,temp;
    double fx;

    if (cam==NULL) return -1;
    if (cam->img==NULL) return -1;

    angle=hangle/2.0;
    temp=tan(angle*M_PI/180.0);

    if (fabs(temp)<1e-8) return -1;

    fx=cam->img->xs/(2.0*temp);
    camSetInternalParams(cam,-fx,-fx,0.0,cam->img->xs/2.0,cam->img->ys/2.0);

    return 0;
}

static int camUndistort(camera *cam,vector3 pcam)
{
    double radius2;

    if (cam==NULL) return -1;

    radius2=pcam[0]*pcam[0]+pcam[1]*pcam[1];
    pcam[0]*=(1+radius2*cam->k1+radius2*radius2*cam->k2);
    pcam[1]*=(1+radius2*cam->k1+radius2*radius2*cam->k2);

    return 0;
}

int camImageToCamera (camera *cam,vector3 pim,vector3 pcam)
{
    if (cam==NULL) return -1;

    //camOriginAtCenter(cam,ptemp);
	mat3MultiplyVector(cam->Kinv,pim,pcam);
    camUndistort(cam,pcam);

	return 0;
}

int camCameraToImage (camera *cam,vector3 pcam,vector3 pim)
{
    if (cam==NULL) return -1;

	mat3MultiplyVector(cam->K,pcam,pim);

	return 0;
}

int camCameraToWorld (camera *cam,double x,double y,double z,double depth,vector4 pworld)
{
    vector4 pcam;
    if (cam==NULL) return -1;
    if (fabs(depth)<1e-8) return -1;

	pcam[0]=x;
	pcam[1]=y;
	pcam[2]=z;
    pcam[3]=1/depth;
	vect4Homogenize3D(pcam);
	mat4MultiplyVector(cam->Minv,pcam,pworld);

	return 0;
}

int camWorldToCamera (camera *cam,vector4 pworld,vector4 pcam)
{
    if (cam==NULL) return -1;

	mat4MultiplyVector(cam->M,pworld,pcam);
	vect4Homogenize3D(pcam);
	vect3Homogenize2D(pcam);

	return 0;
}

int camInFront(camera *cam,vector4 pworld)
{
    vector4 pcam;

    if (cam==NULL) return -1;

	mat4MultiplyVector(cam->M,pworld,pcam);
	vect4Homogenize3D(pcam);

    if (pcam[2]>0) return 1;
    else return 0;
}

int camImageToWorld (camera *cam,double x,double y,double depth,vector4 pworld)
{
    vector4 pcam;
    vector3 pim;
    if (cam==NULL) return -1;
    if (fabs(depth)<1e-8) return -1;

	pim[0]=x;
	pim[1]=y;
	pim[2]=1.0;

    //camOriginAtCenter(cam,pim);
	mat3MultiplyVector(cam->Kinv,pim,pcam);
    camUndistort(cam,pcam);
    pcam[3]=1/depth;
	vect4Homogenize3D(pcam);
	mat4MultiplyVector(cam->Minv,pcam,pworld);

	return 0;
}

int camWorldToImage (camera *cam,vector4 pworld,vector3 pim)
{
    vector4 pcam;
    if (cam==NULL) return -1;

	mat4MultiplyVector(cam->M,pworld,pcam);
	vect4Homogenize2D(pcam);
	mat3MultiplyVector(cam->K,pcam,pim);
    //camOriginAtUpperLeft(cam,pim);

    return 0;
}

int camImageUndistort(camera *cam,vector3 pixel)
{
    vector3 pcam;

    if (cam==NULL) return -1;

    camImageToCamera(cam,pixel,pcam);
    camCameraToImage(cam,pcam,pixel);

    return 0;
}

int camOriginAtCenter(camera *cam,vector3 pixel)
{
  if (cam==NULL || cam->img==NULL) return -1;

  pixel[0]-=cam->img->xs/2.0;
  pixel[1]-=cam->img->ys/2.0;

  return 0;
}

int camOriginAtUpperLeft(camera *cam,vector3 pixel)
{
  if (cam==NULL || cam->img==NULL) return -1;

  pixel[0]+=cam->img->xs/2.0;
  pixel[1]+=cam->img->ys/2.0;

  return 0;
}

static int intersect_lines(vector3 p1,vector3 p2,vector3 q1,vector3 q2,vector4 intersection)
{
  vector3 v1,v2,v3;
  vector3 c1,c2;
  vector3 ac2,ac2c1,mac2c1;
  vector3 av1;
  double a,denom;

  vectxSubtract(p2,p1,v1,3);
  vectxSubtract(q2,q1,v2,3);
  vectxSubtract(q1,p1,v3,3);
  vect3Cross(v3,v2,c1);
  vect3Cross(v1,v2,c2);
  denom=vectxNorm(c2,3);
  if (fabs(denom)<1e-8) return -1;
  a=vectxNorm(c1,3)/denom;
  vectxScale(c2,a,ac2,3);
  vectxSubtract(ac2,c1,ac2c1,3);
  vectxScale(c2,-a,ac2,3);
  vectxSubtract(ac2,c1,mac2c1,3);
  if (vectxNorm(ac2c1,3)>vectxNorm(mac2c1,3)) a=-a;
  vectxScale(v1,a,av1,3);
  vectxAdd(p1,av1,intersection,3);
  intersection[3]=1.0;

  return 0;
}


static int camTriangulate(camera *cam1,camera *cam2,matrix **pts_3d,unsigned char space,unsigned char force_in_front)
{
    int i,j,k,N;
    int ret1,ret2;
    vector4 p1,p2,pcam1,pcam2;
    vector4 pworld1,pworld2;

    if (cam1==NULL || cam2==NULL) return -1;
    if (pts_3d==NULL) return -1;
    if (cam1->features==NULL || cam2->features==NULL) return -1;
    if (cam1->features->cs!=cam2->features->cs) return -1;
    if (space==IMAGE_SPACE)
    {
      if (cam1->features->rs<2) return -1;
      if (cam2->features->rs<2) return -1;
    }
    else
    {
      if (cam1->features->rs<3) return -1;
      if (cam2->features->rs<3) return -1;
    }

    N=cam1->features->cs;
    matAllocate(pts_3d,N,4);

    for (j=0;j<N;j++)
    {
      if (space==IMAGE_SPACE)
      {
        camImageToWorld(cam1,matGet(cam1->features,j,0),matGet(cam1->features,j,1),1.0,p1);
        camImageToWorld(cam2,matGet(cam2->features,j,0),matGet(cam2->features,j,1),1.0,p2);
      }
      else
      {
        camCameraToWorld(cam1,matGet(cam1->features,j,0),matGet(cam1->features,j,1),matGet(cam1->features,j,2),1.0,p1);
        camCameraToWorld(cam2,matGet(cam2->features,j,0),matGet(cam2->features,j,1),matGet(cam2->features,j,2),1.0,p2);
      }
      ret1=intersect_lines(cam1->position,p1,cam2->position,p2,pworld1);
      ret2=intersect_lines(cam2->position,p2,cam1->position,p1,pworld2);
      mat4MultiplyVector(cam1->M,pworld1,pcam1);
      mat4MultiplyVector(cam2->M,pworld2,pcam2);
      if (force_in_front && (ret1!=0 || ret2!=0 || pcam1[2]<0 || pcam2[2]<0))
      //if (ret1!=0 || pcam1[2]<0)
      {
        //put point at infinity w.r.t. cam1 and cam2 and take the average
        if (space==IMAGE_SPACE)
        {
          camImageToWorld(cam1,matGet(cam1->features,j,0),matGet(cam1->features,j,1),1000000.0,pworld1);
          camImageToWorld(cam2,matGet(cam2->features,j,0),matGet(cam2->features,j,1),1000000.0,pworld2);                
        }
        else
        {
          camCameraToWorld(cam1,matGet(cam1->features,j,0),matGet(cam1->features,j,1),matGet(cam1->features,j,2),1000000.0,pworld1);
          camCameraToWorld(cam2,matGet(cam2->features,j,0),matGet(cam2->features,j,1),matGet(cam2->features,j,2),1000000.0,pworld2);                
        }
      }
      for (k=0;k<4;k++)
      {
        //matSet((*pts_3d),j,k,pworld1[k]);
        matSet((*pts_3d),j,k,(pworld1[k]+pworld2[k])/2.0);
      }
    }

    return 0;
}

int camTriangulatePixels(camera *cam1,camera *cam2,matrix **pts_3d,unsigned char force_in_front)
{
  return camTriangulate(cam1,cam2,pts_3d,IMAGE_SPACE,force_in_front);
}

int camTriangulatePoints(camera *cam1,camera *cam2,matrix **pts_3d,unsigned char force_in_front)
{
  return camTriangulate(cam1,cam2,pts_3d,CAMERA_SPACE,force_in_front);
}

int camSelectPose(camera *cam1,camera *cam2,matrix4 *P)
{
  int i,j,maxi;
  int count[4];
  vector4 cpoint1,cpoint2;
  matrix4 I,Mext1cpy,Mext2cpy;
  matrix *wpoints;

  if (cam1==NULL || cam2==NULL) return -1;
  if (cam1==cam2) return -1;

  wpoints=NULL;

  mat4Identity(I);
  mat4Copy(cam1->M,Mext1cpy); //backup external params, they will be copied back at the end of this function
  mat4Copy(cam2->M,Mext2cpy);

  for (i=0;i<4;i++) 
  {
//mat4Print(P[i]);
    count[i]=0;
    camSetExternalParamsFromMatrix(cam1,I);
    camSetExternalParamsFromMatrix(cam2,P[i]);
    camTriangulate(cam1,cam2,&wpoints,IMAGE_SPACE,0);
    if (wpoints!=NULL)
    {
      for (j=0;j<wpoints->cs;j++)
      {
        mat4MultiplyVector(cam1->M,&(wpoints->values[j*wpoints->rs]),cpoint1);
        vect4Homogenize3D(cpoint1);
        mat4MultiplyVector(cam2->M,&(wpoints->values[j*wpoints->rs]),cpoint2);
        vect4Homogenize3D(cpoint2);
        if (cpoint1[2]>0 && cpoint2[2]>0)
        {
          count[i]++;
        }
      }
    }
  }

  maxi=0;
  if (count[1]>count[maxi]) maxi=1;
  if (count[2]>count[maxi]) maxi=2;
  if (count[3]>count[maxi]) maxi=3;
//printf("%d %d %d %d\n",count[0],count[1],count[2],count[3]);

  camSetExternalParamsFromMatrix(cam1,Mext1cpy);
  camSetExternalParamsFromMatrix(cam2,Mext2cpy);

  matFree(&wpoints);

	return maxi;
}

