/*
 * This file is part of IMGU.
 * 
 * @Copyright 2008 Université de Montréal, Laboratoire Vision3D
 *   Sébastien Roy (roys@iro.umontreal.ca)
 *   Vincent Chapdelaine-Couture (chapdelv@iro.umontreal.ca)
 *   Lucie Bélanger
 *   Jamil Draréni 
 *
 * IMGU is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * IMGU is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with IMGU.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "imgu_core.h"

#define INIT_NB_TEXT	20
#define INIT_TEXT_STEP	40

int InternalFormat = Y_ORIGIN_IS_AT_TOP;
int ExternalFormat = Y_ORIGIN_IS_AT_TOP;

/**
 * Allocates memory to store text information.
 *
 * @attention This function keeps any text information that @p I already has. To reset text information, use imguResetText().
 *
 * @attention This function should only be used to allocate an empty image with text information. Otherwise, this function should not be used since
 * memory for text information is always allocated using imguAllocate() or imguLoad().
 *
 * @param[in,out] I Address of an image pointer.
 * @return 0: success; <0: error
 *
 */
static int imguAllocateText(imgu **I) {
	int i;
	char **key, **txt;
	int *txtsz;

	if (I == NULL)
		return -1;

	if ((*I) == NULL) {
		printf("AllocateText on an empty image is not permitted anymore. Allocate first.\n");
		return (-1);
		/***
		 (*I)=(imgu *)(malloc(sizeof(imgu)));
		 if ((*I)==NULL) return -1;
		 ***/
	} else {
		if ((*I)->key && (*I)->txt && (*I)->txtsz) //reuse text space
		{
			//imguResetText((*I));
			return 0;
		}
	}

	if ((*I)->ts == 0) {
		key = (char **) malloc(INIT_NB_TEXT * sizeof(char *));
		if (!key) {
			imguFree(I);
			return -1;
		}
		txt = (char **) malloc(INIT_NB_TEXT * sizeof(char *));
		if (!txt) {
			free(key);
			imguFree(I);
			return -1;
		}
		txtsz = (int *) malloc(INIT_NB_TEXT * sizeof(int));
		if (!txtsz) {
			free(key);
			free(txt);
			imguFree(I);
			return -1;
		}

		(*I)->ts = INIT_NB_TEXT;
		(*I)->key = key;
		(*I)->txt = txt;
		(*I)->txtsz = txtsz;
		for (i = 0; i < (*I)->ts; i++) {
			(*I)->key[i] = (*I)->txt[i] = NULL;
			(*I)->txtsz[i] = 0;
		}
	} else {
		// verify that we have valid pointers...
		if ((*I)->key == NULL || (*I)->txt == NULL || (*I)->txtsz == NULL) {
			printf("WARNING!!! key,txt,txtsz are NULL but ts is not 0!!!!!!!!!!\n");
			exit(0);
		}
	}

	return 0;
}

/**
 * Print some information about a single image
 *
 * @param[in] I Image
 */
void imguDump(imgu *I) {
	int i;
	if (I == NULL)
		return;
	printf("------ Image %d x %d x %d (%d text space) -----", I->xs, I->ys, I->cs, I->ts);
	if (I->data == NULL)
		printf("(no data)");
	if (I->complex == NULL)
		printf("(no complex)");
	printf("\n");

	for (i = 0; i < I->ts; i++) {
		if (I->key[i] == NULL && I->txt[i] == NULL)
			continue;
		if (I->key[i] == NULL || I->txt[i] == NULL) {
			printf("TXT[%d]: Single NULL : illegal.\n", i);
			continue;
		}
		printf("Txt[%s]: %s\n", I->key[i], I->txt[i]);
	}

#ifdef LONG_DEBUG
	int x,y,c;
	for(y=0;y<I->ys;y++) {
		printf("y=%3d: ",y);
		for(x=0;x<I->xs;x++) {
			printf("[");
			for(c=0;c<I->cs;c++) {
				printf("%d",I->data[(y*I->xs+x)*I->cs+c]);
				if( c<I->cs-1 ) printf(",");
			}
			printf("] ");
		}
		printf("\n");
	}
#endif

	printf("---------------------------\n");
}

/**
 * Print some information about a linked list of images
 *
 * @param[in] I Image
 */
void imguDumpMulti(imgu *I) {
	while (I) {
		imguDump(I);
		I = I->next;
	}
	printf("========\n");
}

/**
 * Allocates text and data memory, with extra shorts at the end of data space.
 *
 * Image data is allocate and NOT initialised to 0. 
 *
 * @attention This function keeps any text information that @p I already has. To reset text information, use imguResetText().
 * @warning Any complex information should be assumed lost, except if @p xs,ys,cs correspond exactly to the complex dimensions.
 *
 * @param[in,out] I Address of an image pointer.
 * @param[in] xs Width of image
 * @param[in] ys Height of image
 * @param[in] cs Number of color channels
 * @param[in] extra Number of extra short to add to the buffer (not bytes, shorts!)
 * @return 0: success; <0: error
 *
 */
int imguAllocateExtra(imgu **I, int xs, int ys, int cs, int extra) {
	int size = xs * ys * cs + extra;
	if (imguAllocate(I, size, 1, 1))
		return -1;
	return imguAllocate(I, xs, ys, cs);
}

/**
 * Allocates extra memory after an image, and preserve the image in all cases.
 *
 * @param[in,out] I Address of an image pointer.
 * @param[in] extra Number of extra short to add to the buffer (not bytes, shorts!)
 * @return 0: success; <0: error
 *
 */
int imguExtendExtra(imgu *I, int extra) {
	int sz;
	pix_t *data;

	if (I == NULL)
		return -1;
	if (I->data == NULL)
		return (-1);

	sz = I->xs * I->ys * I->cs + extra;

	if (sz <= I->ds) {
		//printf("imguExtendExtra nothing to do\n");
		return 0; // nothing to do
	}
	pix_t *odata = I->data;

	data = (pix_t *) malloc(sz * sizeof(pix_t));
	if (data == NULL) {
		return -1;
	}

	I->data = data;
	I->ds = sz;

	memcpy(I->data, odata, I->xs * I->ys * I->cs * sizeof(pix_t));

	free(odata);

	return 0;
}

/**
 * Allocates text and data memory.
 *
 * Image data is allocated and NOT initialized to 0. 
 *
 * @attention This function keeps any text information that @p I already has. To reset text information, use imguResetText().
 * @warning Any complex information should be assumed lost, except if @p xs,ys,cs correspond exactly to the complex dimensions.
 * @attention The image is not cleared. (@see imguClear)
 *
 * @param[in,out] I Address of an image pointer.
 * @param[in] xs Width of image
 * @param[in] ys Height of image
 * @param[in] cs Number of color channels
 * @return 0: success; <0: error
 *
 */
int imguAllocate(imgu **I, int xs, int ys, int cs) {
	int sz;
	pix_t *data;

	if (I == NULL)
		return -1;

	if ((*I) == NULL) {
		// reserve new space for the structure itself.
		(*I) = (imgu *) (malloc(sizeof(imgu)));
		if ((*I) == NULL) {
			printf("OutOfMem in imguAllocate\n");
			return -1;
		}

		// initialize some fields
		(*I)->xs = (*I)->ys = (*I)->cs = 0;
		(*I)->ds = 0;
		(*I)->ts = 0;
		(*I)->data = NULL;
		(*I)->complex = NULL;
		(*I)->key = NULL;
		(*I)->txt = NULL;
		(*I)->txtsz = NULL;
		(*I)->next = NULL;
#ifdef HAVE_RQUEUE
		(*I)->Qrecycle=NULL;
#endif
	}

	if (imguAllocateText(I))
		return -1;

	sz = xs * ys * cs;

	if (sz <= (*I)->ds && (*I)->data != NULL) {
		(*I)->xs = xs;
		(*I)->ys = ys;
		(*I)->cs = cs;
		return 0;
	}
	if (sz > (*I)->ds)
		imguFreeComplex(*I);
	else if ((*I)->complex != NULL)
		sz = (*I)->ds; //allocate possibly more memory to be consistent with complex memory (although complex info is probably useless because of dimension change)

	imguFreeData(*I);

	data = (pix_t *) malloc(sz * sizeof(pix_t));
	if (data == NULL) {
		imguFree(I);
		return -1;
	}

	(*I)->xs = xs;
	(*I)->ys = ys;
	(*I)->cs = cs;
	(*I)->data = data;
	(*I)->ds = sz;

	return 0;
}

/**
 * Sets image data to 0.
 *
 */
int imguClear(imgu *I) {
	int sz, i;

	if (I == NULL || I->data == NULL)
		return -1;

	sz = I->xs * I->ys * I->cs;

	for (i = 0; i < sz; i++) {
		I->data[i] = 0;
	}

	return 0;
}

/**
 * Allocates text and data memory.
 *
 * Image data is allocate and NOT initialised to 0. 
 *
 * @attention This function keeps any text information that @p I already has. To reset text information, use imguResetText().
 * @warning Any complex information should be assumed lost, except if @p xs,ys,cs correspond exactly to the complex dimensions.
 *
 * @param[in] xs Width of image
 * @param[in] ys Height of image
 * @param[in] cs Number of color channels
 * @return Pointer to a new imgu image, NULL if there was a problem
 *
 */
imgu *imguCreate(int xs, int ys, int cs) {
	imgu *I = NULL;
	int k;
	k = imguAllocate(&I, xs, ys, cs);
	if (k) {
		imguFree(&I);
		return (NULL);
	}
	return (I);
}

// make sure there is enough space to add n texts
static int imguAllocateMoreText(imgu *I, int n) {
	int newTs;
	char **newKey;
	char **newTxt;
	int *newTxtsz;
	int i;

	if (I == NULL)
		return -1;

	// compte le nb de textes disponibles si >=n, OK!
	for (i = 0; n > 0 && i < I->ts; i++) {
		if (I->key[i] == NULL)
			n--;
	}

	if (n <= 0)
		return (0); // assez de texte!

	// we need 'n' more textes

	//printf("Allocating %d more\n",n);

	newTs = I->ts + n + INIT_TEXT_STEP;
	newKey = (char **) malloc(newTs * sizeof(char *));
	newTxt = (char **) malloc(newTs * sizeof(char *));
	newTxtsz = (int *) malloc(newTs * sizeof(int));

	if (newKey == NULL || newTxt == NULL || newTxtsz == NULL) {
		if (newKey)
			free(newKey);
		if (newTxt)
			free(newTxt);
		if (newTxtsz)
			free(newTxtsz);
		return (-1);
	}

	// copy!!
	for (i = 0; i < I->ts; i++) {
		newKey[i] = I->key[i];
		newTxt[i] = I->txt[i];
		newTxtsz[i] = I->txtsz[i];
	}
	for (; i < newTs; i++) {
		newKey[i] = newTxt[i] = NULL;
		newTxtsz[i] = 0;
	}

	// move data!
	free(I->key);
	free(I->txt);
	free(I->txtsz);

	I->key = newKey;
	I->txt = newTxt;
	I->txtsz = newTxtsz;

	I->ts = newTs;

	return (0);

}

// place le key/txt a l'index i
// free le txt ou cle si necessaire
// si key est NULL, preserve la cle
// si txt est NULL, preserve le txt
static int imguSetText(imgu *I, const char *key, const char *txt, int i) {
	if (I == NULL)
		return -1;

	//printf("Text set [%d] as key=>%s< txt=>%s<\n",i,key,txt);
	if (i < 0 || i >= I->ts)
		return (-1);
	if (key == NULL && I->key[i] == NULL)
		return (-1);
	if (txt == NULL && I->txt[i] == NULL)
		return (-1);

	// reset key and txt
	if (key != NULL && I->key[i] != NULL) {
		free(I->key[i]);
		I->key[i] = NULL;
	}
	if (txt != NULL && I->txt[i] != NULL) {
		free(I->txt[i]);
		I->txt[i] = NULL;
		I->txtsz[i] = 0;
	}

	// si la cle est nulle, on remplace!
	if (I->key[i] == NULL) {
		I->key[i] = (char *) malloc((strlen(key) + 1) * sizeof(char));
		if (I->key[i])
			strcpy(I->key[i], key);
	}

	// si le texte est null, on remplace!
	if (I->txt[i] == NULL) {
		I->txtsz[i] = strlen(txt) + 1;
		I->txt[i] = (char *) malloc(I->txtsz[i] * sizeof(char)); // taille exacte
		if (I->txt[i])
			strcpy(I->txt[i], txt);
	}

	// OOM?
	if (I->key[i] == NULL || I->txt[i] == NULL) {
		if (I->key[i]) {
			free(I->key[i]);
			I->key[i] = NULL;
		}
		if (I->txt[i]) {
			free(I->txt[i]);
			I->txt[i] = NULL;
			I->txtsz[i] = 0;
		}
		return (-1); // OOM!
	}

	//printf("[%d] : key='%s' txt='%s' txtsz=%d\n",i,I->key[i],I->txt[i],I->txtsz[i]);
	return (0);

}

/**
 * Copy text information.
 *
 * @param[out] dest Address of destination image.
 * @param[in] src Pointer to source image.
 * @return 0: success; <0: error
 *
 */
int imguCopyText(imgu *dest, imgu *src) {
	int i;

	if (dest == NULL)
		return -1;
	if (src == NULL)
		return -1;

	if (dest == src)
		return 0;

	//if( imguAllocateText(dest) ) return -1;

	imguResetText(dest);
	imguAllocateMoreText(dest, src->ts);

	for (i = 0; i < src->ts; i++) {
		imguSetText(dest, src->key[i], src->txt[i], i);
	}

	return 0;
}

/**
 * Copy text, data and complex information.
 *
 * @param[out] dest Address of destination image.
 * @param[in] src Pointer to source image.
 * @return 0: success; <0: error
 *
 */
int imguCopy(imgu **dest, imgu *src) {
	int size;
	int i;

	if (dest == NULL)
		return -1;
	if (src == NULL)
		return -1;

	if (*dest == src)
		return 0;

	size = src->xs * src->ys * src->cs;
	if (src->data != NULL) {
		imguAllocate(dest, src->xs, src->ys, src->cs);
		memcpy((*dest)->data, src->data, size * sizeof(pix_t));
	}
	if (src->complex != NULL) {
		imguAllocateComplex(dest, src->xs, src->ys, src->cs);
		for (i = 0; i < size; i++) {
			(*dest)->complex[i][0] = src->complex[i][0];
			(*dest)->complex[i][1] = src->complex[i][1];
		}
	}
	imguCopyText(*dest, src);

	(*dest)->next = NULL;

	return 0;
}

/**
 * Copy text, data and complex information of a list of images.
 *
 * @attention If @p dest is already part of a list of images, then subsequent images in the list are reused (i.e. changed) and new images are created
 * if required. If the number of images pointed by @p dest is larger than the number of images pointed by @p src, then remaining images in the @p dest list are freed.
 *
 * @param[out] dest Address of destination image.
 * @param[in] src Pointer to source image.
 * @return 0: success; <0: error
 *
 */
int imguCopyMulti(imgu **dest, imgu *src) {
	imgu **currdest, *currsrc;
	imgu *dest_next;

	if (dest == NULL)
		return -1;
	if (src == NULL)
		return -1;

	if (*dest == src)
		return 0;

	currdest = dest;
	currsrc = src;
	while (currsrc != NULL) {
		dest_next = NULL;
		if ((*currdest) != NULL)
			dest_next = (*currdest)->next;
		imguCopy(currdest, currsrc);
		(*currdest)->next = dest_next;
		currsrc = currsrc->next;
		currdest = &((*currdest)->next);
	}

	imguFreeMulti(currdest); //in case dest already contained a longer list

	return 0;
}

// retourne l'index de la permiere cle trouvee a partir de la position start
// ou -1 si rien trouve
static int imguFindKey(imgu *I, const char *key, int start) {
	int i;
	if (I == NULL)
		return -1;

	for (i = start; i < I->ts; i++) {
		if (I->key[i] == NULL)
			continue;
		if (strcmp(I->key[i], key) == 0)
			return (i);
	}
	return (-1);
}

/**
 * Add a copy of the (key,txt) pair.
 *
 * The text @p txt will be added even if this key is already defined.
 *
 * @param I imgu image
 * @param key Key string. Must be non null and a regular keyword. See @ref reservedkeys
 * @param txt Text string. Must ne non null.
 * @return 0 on success; -1,-2,-3 on failure
 *
 */
int imguAddText(imgu *I, const char *key, const char *txt) {
	int i;
	if (I == NULL)
		return -1;

	//printf("adding key='%s' txt='%s'\n",key,txt);
	if (imguAllocateMoreText(I, 1))
		return (-1); // unable to add more text

	// trouve un txt vide (on est sur d'en trouver un)
	for (i = 0; i < I->ts; i++) {
		if (I->key[i] == NULL)
			break;
	}
	if (i == I->ts)
		return (-2); // impossible

	if (imguSetText(I, key, txt, i))
		return (-3);

	return (0);
}

/**
 * Append text to a key
 *
 * @param I
 * 	imgu Image
 * @param key
 * 	key string. Must be non null and a regular keyword. See @ref reservedkeys
 * @param txt
 * 	text string. Must ne non null.
 *
 * Add the text @p txt at the end of the @p key value,
 * or create a new text if @p key does not exists.
 *
 */
int imguAppendText(imgu *I, const char *key, const char *txt) {
	int i, LT, LnT, nL;
	char *ntxt;
	if (I == NULL)
		return -1;

	i = imguFindKey(I, key, 0);
	if (i < 0 || I->txt[i] == NULL)
		return (imguAddText(I, key, txt));

	LT = strlen(I->txt[i]); // texte actuel
	LnT = strlen(txt); // texte a ajouter

	// est-ce que la longueur actuelle est suffisante?
	if (I->txtsz[i] - LT > LnT) {
		// on a assez de place!
		strcpy(I->txt[i] + LT, txt);
		return (0);
	}

	// non! On doit allouer +, disons 2x plus...
	nL = (LT + LnT + 1) * 2;
	ntxt = (char *) malloc(nL);
	//printf("malloc %9d bytes\n",nL);

	strcpy(ntxt, I->txt[i]);
	strcpy(ntxt + LT, txt);

	free(I->txt[i]);
	I->txt[i] = ntxt;
	I->txtsz[i] = nL;
	return (0);
}

/** Push text
 *
 * ajoute au debut du texte courant, sinon creer un nouveau texte
 */
int imguPushText(imgu *I, const char *key, const char *txt) {
	int i, LT, LnT, nL;
	char *ntxt;
	if (I == NULL)
		return -1;

	i = imguFindKey(I, key, 0);
	if (i < 0 || I->txt[i] == NULL)
		return (imguAddText(I, key, txt));

	LT = strlen(I->txt[i]); // texte actuel
	LnT = strlen(txt); // texte a ajouter

	// est-ce que la longueur actuelle est suffisante?
	if (I->txtsz[i] - LT > LnT) {
		// on a assez de place!
		nL = (LT + LnT + 1);
		ntxt = (char *) malloc(nL);
		//printf("malloc %9d bytes\n",nL);

		strcpy(ntxt, txt);
		strcpy(ntxt + LnT, I->txt[i]);

		free(I->txt[i]);
		I->txt[i] = ntxt;
		I->txtsz[i] = nL;

		return (0);
	}

	// non! On doit allouer +, disons 2x plus...
	nL = (LT + LnT + 1) * 2;
	ntxt = (char *) malloc(nL);
	//printf("malloc %9d bytes\n",nL);

	strcpy(ntxt, txt);
	strcpy(ntxt + LnT, I->txt[i]);

	free(I->txt[i]);
	I->txt[i] = ntxt;
	I->txtsz[i] = nL;

	return (0);
}

// remplace le texte de la cle key par txt
// Si la cle n'est pas trouvee, ajoute la cle
// Key et Txt sont obligatoires et non NULL
// remplace la permiere cle trouvee (si cle multiple)
int imguReplaceAddText(imgu *I, const char *key, const char *txt) {
	int i;
	if (I == NULL)
		return -1;

	i = imguFindKey(I, key, 0);
	if (i < 0)
		return (imguAddText(I, key, txt));

	if (imguSetText(I, NULL, txt, i))
		return (-1);

	return (0);
}

// remplace le texte de la cle key par txt
// Si la cle n'est pas trouvee, ajoute la cle
// Key et Txt sont obligatoires et non NULL
// remplace la permiere cle trouvee (si cle multiple)
int imguReplaceText(imgu *I, const char *key, const char *txt) {
	int i;
	if (I == NULL)
		return -1;

	i = imguFindKey(I, key, 0);
	if (i < 0)
		return (-1);

	if (imguSetText(I, NULL, txt, i))
		return (-1);
	return (0);
}

// elimine le texte associé à une clé (all=1 : tout, sinon le premier seulement)
// retourne le nb de cle eliminees
int imguRemoveKey(imgu *I, const char *key, int all) {
	int i, n;
	n = 0;
	i = 0;
	do {
		i = imguFindKey(I, key, i);
		if (i < 0)
			return (n);
		// kill key
		if (I->txt[i] != NULL) {
			free(I->txt[i]);
			I->txt[i] = NULL;
			I->txtsz[i] = 0;
		}
		if (I->key[i] != NULL) {
			free(I->key[i]);
			I->key[i] = NULL;
		}
		n++;
		i++;
	} while (all);
	return (n);
}

/**
 * return the text associated with a specific key
 * return a pointer to the text. PLEASE DO NOT free() THE TEXT!!!
 *
 * return NULL if key does not exist
 */
char *imguGetText(imgu *I, const char *key) {
	int i;
	if (I == NULL)
		return NULL;

	i = imguFindKey(I, key, 0);
	if (i < 0)
		return (NULL);
	return (I->txt[i]);
}

/**
 * Get all key-text pairs associated to a specific key
 *
 * @return number of key-text pairs for the key
 *
 * Fill the table TxtTab up to max values. TxtTab[] must exist for [0..max-1]
 */
int imguGetAllText(imgu *I, const char *key, char **TxtTab, int max) {
	int i, j;
	if (I == NULL)
		return -1;

	i = 0;
	j = 0;
	for (j = 0; j < max; j++) {
		i = imguFindKey(I, key, i);
		if (i < 0)
			break;
		TxtTab[j] = I->txt[i];
		i++;
	}
	return (j);
}

/**
 * Free image text, data and complex information without freeing the image structure.
 *
 * Reset all data and text associated with an image. Also reset the link of the image. Does not free the @c imgu structure itself.
 *
 * @param I Image
 *
 */
void imguReset(imgu *I) {
	if (I == NULL)
		return;

	if (I->data) {
		free(I->data);
		I->data = NULL;
	}
	if (I->complex) {
		free(I->complex);
		I->complex = NULL;
	}

	imguResetText(I);
	/* free the text buffer itself */
	if (I->key) {
		free(I->key);
		I->key = NULL;
	}
	if (I->txt) {
		free(I->txt);
		I->txt = NULL;
	}
	if (I->txtsz) {
		free(I->txtsz);
		I->txtsz = NULL;
	}
	I->xs = I->ys = I->cs = I->ds = 0;
	I->ts = 0;
	I->next = NULL;
}

/**
 * Reset all text information associated with an image.
 *
 * @param[in] I Image
 */
void imguResetText(imgu *I) {
	int i;
	if (I == NULL)
		return;

	for (i = 0; i < I->ts; i++) {
		if (I->key[i]) {
			free(I->key[i]);
			I->key[i] = NULL;
		}
		if (I->txt[i]) {
			free(I->txt[i]);
			I->txt[i] = NULL;
			I->txtsz[i] = 0;
		}
	}
}

/**
 * Adds image @p I to the end of image list @p head
 *
 * @param[out] head Adress of image list
 * @param[in] I Image
 */
int imguAddLastMulti(imgu **head, imgu *I) {
	imgu *currimg;

	if (head == NULL)
		return -1;
	if (I == NULL)
		return -1;

	if (*head == I)
		return 0;

	if ((*head) == NULL) {
		(*head) = I;
		return 0;
	}

	currimg = (*head);
	while (currimg->next != NULL) {
		currimg = currimg->next;
	}

	currimg->next = I;

	return 0;
}

/**
 * Adds image @p I to the start of image list @p head
 *
 * @param[out] head Adress of image list
 * @param[in] I Image
 */
int imguAddFirstMulti(imgu **head, imgu *I) {
	if (head == NULL)
		return -1;
	if (I == NULL)
		return -1;

	I->next = (*head);
	(*head) = I;

	return 0;
}

/**
 * Extract image @p I from the start of the image list @p head
 *
 * @param[out] head Adress of image list
 * @param[in] I Image
 */
imgu *imguRemoveFirstMulti(imgu **head) {
	imgu *I;

	if (head == NULL || (*head) == NULL)
		return NULL;
	I = *head;
	*head = (*head)->next;
	I->next = NULL;
	return (I);
}

/**
 * Extract image @p I from the end of the image list @p head
 *
 * @param[out] head Adress of image list
 * @param[in] I Image
 */
imgu *imguRemoveLastMulti(imgu **head) {
	imgu *I, *Iprev;

	if (head == NULL || (*head) == NULL)
		return NULL;
	I = (*head)->next;
	Iprev = (*head);
	if (I == NULL) //list has only one element
	{
		*head = NULL;
		return Iprev;
	}

	while (I->next != NULL) //find last element
	{
		Iprev = I;
		I = I->next;
	}
	Iprev->next = NULL;
	return (I);
}

/**
 * Free an image completely
 *
 * Free all text, data and complex buffers associated with an image.
 * Then free the @c imgu structure itself and sets @p I to NULL.
 *
 * @param I Image pointer (will be changed)
 * @return nothing. But @c *I is set to NULL.
 */
void imguFree(imgu **I) {
	if (I == NULL || (*I) == NULL)
		return;
	imguReset((*I));
	free((*I));
	(*I) = NULL;
}

/**
 * Free completely a list of images
 *
 * see imguFree()
 */
void imguFreeMulti(imgu **I) {
	if (I == NULL)
		return;

	imgu *J;

	while (*I != NULL) {
		// free *I, but save the next link before its too late...
		J = (*I)->next;
		imguFree(I);
		*I = J;
	}
}

/**
 * Allocates image with complex data.
 *
 * An integer image will also be allocated.
 * Complex (and integer) information is only allocated and NOT initialized to 0 (see imguClearComplex() and imguClear()). 
 *
 * @attention This function keeps any text information that @p I already has. To reset text information, use imguResetText().
 * @warning Any data information should be assumed lost, except if @p xs,ys,cs correspond exactly to the data dimensions.
 *
 * @param[out] cd Address of an image pointer.
 * @param xs Width of image
 * @param ys Height of image
 * @param cs Number of color channels
 * @return 0: success; <0: error
 *
 */
int imguAllocateComplex(imgu **cd, int xs, int ys, int cs) {
	fftw_complex *d;
	int Size;

	if (cd == NULL)
		return -1;

	if (imguAllocate(cd, xs, ys, cs))
		return -1;

	Size = xs * ys * cs;

	if (Size <= (*cd)->ds && (*cd)->complex != NULL) {
		(*cd)->xs = xs;
		(*cd)->ys = ys;
		(*cd)->cs = cs;
		//for (i=0;i<Size;i++)
		//{
		//    (*cd)->complex[i][0]=0;
		//    (*cd)->complex[i][1]=0;
		//}
		return 0;
	}
	if (Size > (*cd)->ds)
		imguFreeData(*cd);
	else if ((*cd)->data != NULL)
		Size = (*cd)->ds; //allocate possibly more memory to be consistent with data memory (although data info is probably useless because of dimension change)

	imguFreeComplex(*cd);

	d = (fftw_complex *) (malloc(sizeof(fftw_complex) * Size));

	if (d == NULL) {
		(*cd)->xs = (*cd)->ys = (*cd)->cs = (*cd)->ds = 0;
		(*cd)->complex = NULL;
		return -1;
	}

	(*cd)->xs = xs;
	(*cd)->ys = ys;
	(*cd)->cs = cs;
	(*cd)->ds = Size;
	(*cd)->complex = d;

	//for (i=0;i<Size;i++)
	//{
	//    (*cd)->complex[i][0]=0;
	//    (*cd)->complex[i][1]=0;
	//}

	return 0;
}

/**
 * Sets complex data to 0.
 * 
 * @param I Image to clear
 * @param component Indicates the component (0 for Real, 1 for Imaginary, -1 for both).
 *
 */
int imguClearComplex(imgu *I, unsigned char component) {
	int sz, i;

	if (component > 1)
		return -1;
	if (I == NULL || I->complex == NULL)
		return -1;

	sz = I->xs * I->ys * I->cs;

	for (i = 0; i < sz; i++)
		I->complex[i][component] = 0;

	return 0;
}

/**
 * Free complex data.
 * Only the data is freed. The image structure and any interger data is preserved.

 * @param[in,out] cd Address of an image pointer.
 */
void imguFreeComplex(imgu *I) {
	if (I == NULL)
		return;
	if (I->complex != NULL)
		free(I->complex);
	I->complex = NULL;
	if (I->data == NULL)
		I->xs = I->ys = I->cs = I->ds = 0;
}

void imguFreeData(imgu *I) {
	if (I == NULL)
		return;
	if (I->data != NULL)
		free(I->data);
	I->data = NULL;
	if (I->complex == NULL)
		I->xs = I->ys = I->cs = I->ds = 0;
}

