#ifndef IMGU_CORE_H
#define IMGU_CORE_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * This file is part of IMGU.
 * 
 * @Copyright 2008 Université de Montréal, Laboratoire Vision3D
 *   Sébastien Roy (roys@iro.umontreal.ca)
 *   Vincent Chapdelaine-Couture (chapdelv@iro.umontreal.ca)
 *   Lucie Bélanger
 *   Jamil Draréni 
 *
 * IMGU is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * IMGU is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with IMGU.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef IMGU8
#define pix_t		unsigned char
#define IMGU_MAXVAL 255
#else
#define pix_t		unsigned short
#define IMGU_MAXVAL 65535
#endif

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795029L
#endif

//#define COMPLEX_FLOAT
#ifdef COMPLEX_FLOAT
#define complex_t float
#else
#define complex_t double
#endif

typedef complex_t fftw_complex[2];

typedef struct imgu
{
	int xs; ///< width of image
	int ys; ///< height of image
	int cs; ///< number of channels (1=mono, 2=gray+alpha, 3=rgb, 4=rgba, >4=ok but undefined)
	int ds; ///< number of elements in data array (0 when data==NULL) Can be > xs*ys*cs
	pix_t *data; /**< Image Data: row-order (y,x,c)
	 *  In RGB mode : RGB RGB RGB RGB ...
	 *  NULL -> No Image loaded
	 *  Pixel (x,y,c) is at data[(y*xs+x)*cs+c]
	 */

	fftw_complex *complex; // Complex image. Array is also of size ds

	int ts; ///< nb de textes disponibles
	char **key; ///< [ts] contient des strings avec des cles pour les textes.
	///< @attention certains [i] peuvent etre NULL
	char **txt; ///< [ts] contient des textes qui vont avec les cles
	///< @attention certains [i] peuvent etre NULL
	int *txtsz; ///< [ts] contient la longueur du texte allouee par le malloc.
	///< quand on fait append, on double la taille si on fait malloc. sinon, ordinaire.
	///< txtlen n'a pas a etre mis a 0 lorsque txt[] est null.

	struct imgu *next;
} imgu;

/// acces a un pixel
#define INDEX(I,x,y,c)	(((y)*(I)->xs+(x))*(I)->cs+(c))
#define PIXEL(I,x,y,c)	((I)->data[INDEX(I,x,y,c)])
#define COMPLEX(I,x,y,c,u)	((I)->complex[INDEX(I,x,y,c)][u])

#define Y_ORIGIN_IS_AT_TOP	0
#define Y_ORIGIN_IS_AT_BOTTOM	1

extern int InternalFormat; ///< Where is the Y=0 row in memory?
extern int ExternalFormat; ///< Where is the Y=0 row in external file (for load and save)?

#ifdef __cplusplus
extern "C" {
#endif

	/**
	 * @name Debug and image information
	 * @{
	 */

	void imguDump(imgu *I);
	void imguDumpMulti(imgu *I);

	/*@}*/

	/**
	 * @name Text tags manipulation
	 *
	 * @par
	 * These functions manipulate Text tags associated with image.
	 * Each png image has a list of (key,value) pairs that can be used to add text
	 * information to an image. The format is not strict, except that the key must
	 * be a simple keyword and the text must be a nul-terminated string.
	 * Key and text should be ascii only.
	 *
	 * @par
	 * There is a set of reserved keys. They are in uppercase, so used lower case keys for your own
	 * non reserved keys. The list of reserved tags is documented in @ref reservedkeys.
	 *
	 * @{
	 */

	int imguCopyText(imgu *dest, imgu *src);
	void imguResetText(imgu *I);

	int imguAddText(imgu *I, const char *key, const char *txt);
	int imguAppendText(imgu *I, const char *key, const char *txt);
	int imguPushText(imgu *I, const char *key, const char *txt); ///< ajoute le texte au debut du courant, ou cree
	int imguReplaceAddText(imgu *I, const char *key, const char *txt); ///< replace or add une copie du texte
	int imguReplaceText(imgu *I, const char *key, const char *txt); ///< replace (or fail) une copie du texte
	int imguRemoveKey(imgu *I, const char *key, int all); ///< enleve une cle (all ou single)


	char *imguGetText(imgu *I, const char *key); ///< trouve le texte pour cette cle

	// retourne le nb de textes associes a la cle key
	// rempli le tableau fourni jusqu'a max valeurs
	int imguGetAllText(imgu *I, const char *key, char **TxtTab, int max);

	/*@}*/

	/**
	 * @name Memory Memory management
	 *
	 * @par
	 * These functions allocate data memory for images.
	 *
	 * @attention The imguAllocate() function should only be called to create a new empty image; it does not have to be called before imguLoad().
	 *
	 * @{
	 */

	imgu *imguCreate(int xs, int ys, int cs);
	int imguAllocate(imgu **I, int xs, int ys, int cs);
	int imguAllocateExtra(imgu **I, int xs, int ys, int cs, int extra);
	int imguExtendExtra(imgu *I, int extra); // ajoute extra shorts au bout de l'image, en conservant l'image
	int imguClear(imgu *I);

	int imguCopy(imgu **dest, imgu *src);
	int imguCopyMulti(imgu **dest, imgu *src);

	void imguFree(imgu **I);
	void imguFreeData(imgu *I);
	void imguFreeMulti(imgu **I);

	int imguAddFirstMulti(imgu **head, imgu *I);
	int imguAddLastMulti(imgu **head, imgu *I);
	imgu *imguRemoveFirstMulti(imgu **head);
	imgu *imguRemoveLastMulti(imgu **head);

	/**
	 * Allocates image with complex data.
	 *
	 * An integer image will also be allocated.
	 * Complex (and integer) information is only allocated and NOT initialized to 0 (see imguClearComplex() and imguClear()).
	 *
	 * @attention This function keeps any text information that @p I already has. To reset text information, use imguResetText().
	 * @warning Any data information should be assumed lost, except if @p xs,ys,cs correspond exactly to the data dimensions.
	 *
	 * @param[out] cd Address of an image pointer.
	 * @param xs Width of image
	 * @param ys Height of image
	 * @param cs Number of color channels
	 * @return 0: success; <0: error
	 *
	 */
	int imguAllocateComplex(imgu **cd, int xs, int ys, int cs);
	int imguClearComplex(imgu *I, unsigned char component);
	void imguFreeComplex(imgu *I);
	void imguFreeData(imgu *I);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif

