/*
 * main.cpp
 *
 */

// Standard library imports
#include <iostream>
#include <stdexcept>

// Third party imports
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>

// Local imports
#include "folder.h"
#include "minimize.h"
#include "mosaic.h"

// Namespaces
using namespace cv;
using namespace std;

/*!********************************************************************************
 * Argument reading
 **********************************************************************************/

// Program arguments
struct Arg
{
	Arg() :
		matchRatioThreshold(0.6), matchDistanceThreshold(0.1), height(-1),
		radius(38), blending(true), quiet(false), output("result.png") {
	}

	double matchRatioThreshold;
	double matchDistanceThreshold;
	double radius;

	bool blending;
	bool quiet;

	int height;

	vector<Mat> center;
	vector<Mat> offset;

	string output;
};

// Print usage
void help(char ** argv) {
	cerr << endl;
	cerr << "usage: " << argv[0] << " [<options>] <center> <offset>" << endl;
	cerr << "    -h, --help               : print this help" << endl;
	cerr << "    -t, --threshold <double> : ratio threshold on 2nd best fit to filter matches (default = 0.6)" << endl;
	cerr << "    -d, --distance  <double> : maximal distance for the best fit when matching (default = 0.1)" << endl;
	cerr << "    -s, --size <int>         : height (in pixel) of the projection cylinder (default = input height)" << endl;
	cerr << "    -i, --iter <int>         : maximum number of iteration for bundle adjustment (default = 5000)" << endl;
	cerr << "    -l, --limit <double>     : distance limit for outliers in the bundle adjustment (default = 8)" << endl;
	cerr << "    -r, --radius             : radius of the circle for the shifted camera (default = 38cm)" << endl;
	cerr << "    -a, --assemble           : assemble the images (with no blending)" << endl;
	cerr << "    -q, --quiet              : suppress informations that are supplied on stdout" << endl;
	cerr << "    -o, --output <name>      : filename suffix for the resulting panoramas" << endl;
	cerr << "    <center>                 : directory for the pictures taken from the centered camera" << endl;
	cerr << "    <offset>                 : directory for the pictures taken from the shifted  camera" << endl;
	cerr << endl;
}

// Parse command line arguments
void parse(int argc, char** argv, Arg & args) {
	if (argc < 3) {
		help(argv);
		throw invalid_argument("Wrong number of arguments");
	}

	// Parameters
	int index = 0;
	int argIterMax = -1;
	double argOutliersLimit = -1;

	// Read arguments
	for (int i = 1; i < argc; ++i) {
		if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
			help(argv);
		if (!strcmp(argv[i], "-q") || !strcmp(argv[i], "--quiet"))
			args.quiet = false;
		if (!strcmp(argv[i], "-a") || !strcmp(argv[i], "--assemble"))
			args.blending = false;
		else if ((!strcmp(argv[i], "-o") || !strcmp(argv[i], "--output")) && i + 1 < argc)
			args.output = argv[++i];
		else if ((!strcmp(argv[i], "-t") || !strcmp(argv[i], "--threshold")) && i + 1 < argc)
			args.matchRatioThreshold = strtod(argv[++i], NULL);
		else if ((!strcmp(argv[i], "-d") || !strcmp(argv[i], "--distance")) && i + 1 < argc)
			args.matchDistanceThreshold = strtod(argv[++i], NULL);
		else if ((!strcmp(argv[i], "-r") || !strcmp(argv[i], "--radius")) && i + 1 < argc)
			args.radius = strtod(argv[++i], NULL);
		else if ((!strcmp(argv[i], "-l") || !strcmp(argv[i], "--limit")) && i + 1 < argc)
			argOutliersLimit = strtod(argv[++i], NULL);
		else if ((!strcmp(argv[i], "-s") || !strcmp(argv[i], "--size")) && i + 1 < argc)
			args.height = strtoul(argv[++i], 0, 10);
		else if ((!strcmp(argv[i], "-i") || !strcmp(argv[i], "--iter")) && i + 1 < argc)
			argIterMax = strtoul(argv[++i], 0, 10);
		else
			switch (index++) {
			case 0:
				// Load the centered pictures
				loadFolder(argv[i], args.center);
				break;
			case 1:
				// Load the shifted pictures
				loadFolder(argv[i], args.offset);
				break;
			default:
				throw invalid_argument(argv[i]);
			}
	}

	// Check correctness of arguments
	if (args.matchRatioThreshold < 0 || args.matchRatioThreshold >= 1)
		throw invalid_argument("Wrong value for ratio threshold");

	if (args.matchDistanceThreshold <= 0)
		throw invalid_argument("Wrong value for distance threshold");

	if (args.radius <= 0)
		throw invalid_argument("Wrong circle radius");

	if (args.center.size() != args.offset.size())
		throw invalid_argument("Not the same number of pictures in <center> and <offset>");

	// Check that images have the same size
	Size refCenter = args.center[0].size();
	Size refOffset = args.center[0].size();
	for (int i = 1; i < args.center.size(); ++i) {
		if (args.center[i].size() != refCenter)
			throw invalid_argument("Center images must have the same size");
		if (args.offset[i].size() != refOffset)
			throw invalid_argument("Shifted images must have the same size");
	}

	// Set some parameters
	if (argIterMax > 0)
		setMaxIter(argIterMax);

	if (argOutliersLimit > 0)
		setThreshold(argOutliersLimit);

	if (args.height < 0)
		args.height = args.offset[0].rows;
}

/*!********************************************************************************
 * Main program
 **********************************************************************************/
int main(int argc, char** argv) {
	Arg args;
	parse(argc, argv, args);

	namedWindow("Concentric", CV_WINDOW_NORMAL);

	// Compute the panoramas
	ConcentricMosaic wrapper(args.matchRatioThreshold, args.matchDistanceThreshold, args.quiet);
	wrapper.compute(args.center, args.offset, args.radius, args.height, args.blending, args.output);

	return EXIT_SUCCESS;
}
