/*
 * minimize.cpp
 *
 * Code for non-linear optimization, for two different situations (see the report).
 * To display information about the bundle adjustment during the execution, just
 * #define VERBOSE at compile time
 */

#define VERBOSE

// Standard library imports
#include <iostream>
#include <exception>

// Local imports
#include "imgu/imgu.h"
#include "definition.h"
#include "rotations.h"
#include "minimize.h"

// Namespaces
using namespace cv;
using namespace std;

// Bundle adjustment global variables
extern void *nl_fcn_global_user_data[2];
extern int nl_fcn_global_max_nb_iter;
extern int nl_fcn_global_iter_nb;
extern int nl_fcn_global_options;
extern double *nl_fcn_global_errors;

/*!********************************************************************************
 * Miscellaneous
 **********************************************************************************/

// Program options
static int maxNumberIteration = 5000;
static double sigma = 8;

void setMaxIter(int iterMax) {
	maxNumberIteration = iterMax;
}

void setThreshold(double threshold) {
	sigma = threshold;
}

// A robust error function when computing the reprojection error
inline double errorFunction(const Vec2f & r) {
	double L1 = norm(r, NORM_L1);
	if (L1 < sigma)
		return norm(r, NORM_L2);
	else
		return 2 * sigma * L1 - sigma * sigma;
}

// Check whether the maximum number of iteration during a bundle adjustment is reached
inline bool maxIteration(int nbeqs, double *y) {
	if (nl_fcn_global_iter_nb >= nl_fcn_global_max_nb_iter) {
		for (int k = 0; k < nbeqs; ++k)
			y[k] = 0.0;
		cerr << "Warning: maximum number of iterations reached (NLoptimization)" << endl;
		return true;
	}
	return false;
}

/*!********************************************************************************
 * Non-Linear Optimization for the focal length, based on precomputed homographies
 **********************************************************************************/

int computeErrorFocal(void *p, int nbeqs, int nbvars, const double *x, double *y, int flag) {
	if (maxIteration(nbeqs, y))
		return 0;

	// Retrieve global informations
	vector<Mat> & homography = *(vector<Mat> *) (nl_fcn_global_user_data[0]);
	vector<double> & angle = *(vector<double> *) (nl_fcn_global_user_data[1]);

	double rotation = 0;
	for (int i = 0; i < homography.size(); ++i) {
		// Angle[i] = Rotation angle from img_0 to img_i
		angle[i] = rotation;
		// Assumes H_{i,i+i} = V⋅R, with R a rotation around the axis y = (0,1,0)
		rotation += asin(homography[i].at<double> (0, 2) / *x);
	}

	// The final rotation must be either 2π or -2π
	*y = fabs(2 * Pi - fabs(rotation));

	#ifdef VERBOSE
		cout << "Step: " << nl_fcn_global_iter_nb << ", Focal: " << *x << ", Error: " << *y << endl;
	#endif

	++nl_fcn_global_iter_nb;
	return 0;
}

// Main function for the Focal Length NL-Optimization
void optimizeFocalLength(vector<Mat> & homography, vector<double> & angle, double & focal) {
	matrix * params = NULL;
	matAllocate(&params, 1, 1);
	params->values[0] = focal;
	matNLOptimization(params, 1, maxNumberIteration, &homography, &angle, computeErrorFocal, NULL, 0);
	focal = params->values[0];
	matFree(&params);
}

/*!********************************************************************************
 * Non-Linear Optimization for every parameter, using the reconstruction errors
 **********************************************************************************/

int computeErrorReprojection(void *p, int nbeqs, int nbvars, const double *x, double *y, int flag) {
	if (maxIteration(nbeqs, y))
		return 0;

	// Retrieve global informations
	MatchGraph & graph = *(MatchGraph *) (nl_fcn_global_user_data[0]);

	// Focal length matrix
	double F[3][3] = {
		{ x[0], 0, 1 },
		{ 0, x[0], 1 },
		{ 0, 0, 1 } };
	Mat mFocal = Mat(3, 3, CV_64F, F);

	// Normalized rotation axis
	double len = sqrt(x[1] * x[1] + x[2] * x[2] + x[3] * x[3]);
	Vec3d axis(x[1] / len, x[2] / len, x[3] / len);
	y[0] = fabs(len - 1); // Keep the length close to 1

	// Compute the reprojection error for each pair of keypoints
	int index = 1, n = graph.size();
	Mat acc = mFocal * rotationFromAxis(axis, 0); // theta_0 = 0
	for (int i = 0; i < n; ++i) {
		// x[3 + i] = Rotation angle from img_0 to img_i
		// acc  = Projection matrix for the camera_i
		// next = Projection matrix for the camera_{i+1} (and img_n = img_0)
		Mat next = mFocal * rotationFromAxis(axis, i + 1 == n ? 0 : x[3 + i + 1]);
		Mat residuals;

		// Mapping Left → Right
		perspectiveTransform(graph[i].first, residuals, next * acc.inv());
		residuals -= graph[i].second;
		for (int j = 0; j < residuals.rows; ++j)
			y[index + j] = errorFunction(residuals.at<Vec2f> (j, 0));
		index += residuals.rows;

		// Mapping Right → Left
		perspectiveTransform(graph[i].second, residuals, acc * next.inv());
		residuals -= graph[i].first;
		for (int j = 0; j < residuals.rows; ++j)
			y[index + j] = errorFunction(residuals.at<Vec2f> (j, 0));
		index += residuals.rows;

		swap(acc, next);
	}

	#ifdef VERBOSE
		// Display periodic informations
		if (nl_fcn_global_iter_nb % 25 == 0) {
			double meanError = 0;
			for (int i = 1; i < nbeqs; ++i)
				meanError += y[i];
			meanError /= (nbeqs - 1);
			cout << "Step: " << nl_fcn_global_iter_nb << ", Focal: " << x[0];
			cout << ", Error: " << meanError << ", Axis: " << Mat(axis, false) << endl;
		}
	#endif

	++nl_fcn_global_iter_nb;
	return 0;
}

void optimizeReprojection(MatchGraph & graph, vector<double> & angle, Vec3d & axis, double & focal) {
	matrix * params = NULL;
	matAllocate(&params, 3 + angle.size(), 1);

	// Number of errors computed with the keypoints
	int nsamples = 0;
	for (int i = 0; i < graph.size(); ++i)
		nsamples += graph[i].first.rows;

	// Record initial parameters
	params->values[0] = focal;
	for (int i = 0; i < 3; ++i)
		params->values[i + 1] = axis[i];
	for (int i = 1; i < angle.size(); ++i)
		params->values[i + 3] = angle[i];

	// Optimize the parameters
	matNLOptimization(params, 1 + 2 * nsamples, maxNumberIteration, &graph, NULL, computeErrorReprojection, NULL, 0);

	// Record the results
	focal = params->values[0];
	for (int i = 0; i < 3; ++i)
		axis[i] = params->values[i + 1];
	for (int i = 1; i < angle.size(); ++i)
		angle[i] = params->values[i + 3];

	// Normalize the rotation axis
	double len = norm(axis);
	for (int i = 0; i < 3; ++i)
		axis[i] /= len;

	matFree(&params);
}
