/*
 * mosaic.cpp
 *
 */

// Standard library imports
#include <iostream>
#include <exception>
#include <ctime>

// Third party imports
#include <opencv2/core/core.hpp>
#include <opencv2/features2d/features2d.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>

// Local imports
#include "definition.h"
#include "rotations.h"
#include "minimize.h"
#include "mosaic.h"

// Namespaces
using namespace cv;
using namespace std;

/*!********************************************************************************
 * Custom exceptions
 **********************************************************************************/

class BadKeypoints: public exception
{
	virtual const char* what() const throw () {
		return "Not enough keypoints were found to estimate the homography.";
	}
};

class BadHomography: public exception
{
	virtual const char* what() const throw () {
		return "Found an homography with rank < 3.";
	}
};

class BadSize: public exception
{
	virtual const char* what() const throw () {
		return "Final window size was incorrect. Make sure you have selected the proper pictures.";
	}
};

/*!********************************************************************************
 * Time-related functions
 **********************************************************************************/

// Retrieve current time
inline double getTime() {
	timespec time;
	clock_gettime(CLOCK_MONOTONIC, &time);

	// Return the time in seconds
	return (double) time.tv_sec + time.tv_nsec / 1000000000.0;
}

// Updates the current time and returns the elapsed time in second
inline double getDiffTime(double & t) {
	double newTime = getTime();
	double diffTime = newTime - t;
	t = newTime;
	return diffTime;
}

/*!********************************************************************************
 * Miscellaneous class methods
 **********************************************************************************/

ConcentricMosaic::ConcentricMosaic(double ratio, double distance, bool quiet) :
	matchRatioThreshold(ratio), matchDistanceThreshold(distance), quiet(quiet) {
}

inline void ConcentricMosaic::printTime(const string & message) {
	if (not quiet)
		cout << "***** " << message << "(sec) " << getDiffTime(time) << endl;
}

/*!********************************************************************************
 * Main algorithm
 **********************************************************************************/

void ConcentricMosaic::compute(const vector<Mat> & center, const vector<Mat> & offset,
	double radius, int height, bool blending, const string & output) {
	// The angular deviation to retrieve a ray for each eye
	// (6.5cm = distance between the two eyes)
	double phi = atan((0.5 * 6.5) / radius);

	// Miscellaneous parameters
	double T[3][3] = {
		{ 1, 0, center[0].cols / 2. },
		{ 0, 1, center[0].rows / 2. },
		{ 0, 0, 1 } };
	mTranslate = Mat(3, 3, CV_64F, T).clone();
	time = getTime();
	n = center.size();

	originalWidth = center[0].cols;
	originalHeight = center[0].rows;
	finalHeight = height;

	// Vectors initialization
	graph.resize(n);
	keypoints.resize(n);
	descriptors.resize(n);
	projection.resize(n);
	angle.resize(n);

	// Main algorithm
	findKeypoints(center);
	findMatches();
	initalizeBundleAdj();
	performBundleAdj();
	createWorldMap(center);

	// Stitch the images together
	StitchFunction function = blending ? &ConcentricMosaic::blend : &ConcentricMosaic::assemble;
	int rType = blending ? CV_32FC3 : CV_8UC3;

	Mat finalImage = Mat(finalHeight, finalWidth, rType, Scalar(0, 0, 0));
	for (int i = 0; i < n; ++i)
		(this->*function)(center[i], i, 0, finalImage);
	imwrite("central-" + output, finalImage);
	printTime("Centra panorama created in ");

	finalImage = Scalar(0, 0, 0);
	for (int i = 0; i < n; ++i)
		(this->*function)(offset[i], i, phi, finalImage);
	imwrite("left-" + output, finalImage);
	printTime("Left panorama created in ");

	finalImage = Scalar(0, 0, 0);
	for (int i = 0; i < n; ++i)
		(this->*function)(offset[i], i, -phi, finalImage);
	imwrite("right-" + output, finalImage);
	printTime("Right panorama created in ");
}

/*!********************************************************************************
 * STEP 1: Find SIFT features and match them between the images
 **********************************************************************************/

// SIFT features extraction
void ConcentricMosaic::findKeypoints(const vector<Mat> & img) {
	SurfFeatureDetector detector;
	SurfDescriptorExtractor extractor;

	for (int i = 0; i < (int) img.size(); ++i) {
		// Detecting keypoints
		detector.detect(img[i], keypoints[i]);
		// Computing descriptors
		extractor.compute(img[i], keypoints[i], descriptors[i]);
	}

	printTime("Keypoints found in ");
}

// Find the corresponding matches
void ConcentricMosaic::findMatches() {
	for (int i = 0; i < n; ++i) {
		Mat & descriptors1 = descriptors[i];
		Mat & descriptors2 = descriptors[(i + 1) % n];
		vector<KeyPoint> & keypoints1 = keypoints[i];
		vector<KeyPoint> & keypoints2 = keypoints[(i + 1) % n];

		// Matching descriptors
		FlannBasedMatcher matcher;
		vector<vector<DMatch> > matches;
		matcher.knnMatch(descriptors1, descriptors2, matches, 2);

		// Filter ambiguous matches
		Mat pts1(0, 1, CV_32FC2), pts2(0, 1, CV_32FC2);
		for (int k = 0; k < matches.size(); ++k) {
			DMatch m1 = matches[k][0];
			DMatch m2 = matches[k][1];
			// If the match is unambiguous enough
			if (m1.distance < matchDistanceThreshold && m1.distance < matchRatioThreshold * m2.distance) {
				pts1.push_back<Point2f>(keypoints1[m1.queryIdx].pt);
				pts2.push_back<Point2f>(keypoints2[m1.trainIdx].pt);
			}
		}

		// Check whether enough matching remain
		if (pts1.rows < 4 || pts2.rows < 4)
			throw BadKeypoints();

		// Translate the keypoints so that the image is center in (w/2, h/2)
		perspectiveTransform(pts1, pts1, mTranslate.inv());
		perspectiveTransform(pts2, pts2, mTranslate.inv());

		graph[i].first = pts1;
		graph[i].second = pts2;
	}

	keypoints.clear();
	descriptors.clear();
	printTime("Matches found in ");
}

/*!********************************************************************************
 * STEP 2: Bundle adjustment
 **********************************************************************************/

// First simple bundle adjustment to initialize the focal and rotation angles
void ConcentricMosaic::initalizeBundleAdj() {
	// First estimation of the homographies
	vector<Mat> homography(n);
	for (int i = 0; i < n; ++i) {
		homography[i] = findHomography(graph[i].first, graph[i].second, CV_RANSAC);
		homography[i] /= homography[i].at<double> (1, 1);

		if (fabs(determinant(homography[i])) < fEpsilon)
			throw BadHomography();
	}

	// First guess of the focal length
	vector<double> guess;
	for (int i = 0; i < n; ++i) {
		// A rough but very unstable estimation of the focal length. We suppose
		// that as H = V⋅R, its first two rows (resp. cols) have the same length
		Mat_<double> h = homography[i];
		double f1 = sqrt( (h(1, 2) * h(1, 2) - h(0, 2) * h(0, 2))
			/ (h(0, 0) * h(0, 0) + h(0, 1) * h(0, 1) - h(1, 0) * h(1, 0) - h(1, 1) * h(1, 1)) );
		double f2 = sqrt( (h(0, 0) * h(0, 0) + h(1, 0) * h(1, 0) - h(0, 1) * h(0, 1) - h(1, 1) * h(1, 1))
			/ (h(2, 1) * h(2, 1) - h(2, 0) * h(2, 0)) );

		if (not isnan(f1))
			guess.push_back(f1);
		if (not isnan(f2))
			guess.push_back(f2);
	}
	// We then take the median of all the estimated values
	sort(guess.begin(), guess.end());
	focal = guess[guess.size() / 2];

	// Bundle adjustment
	optimizeFocalLength(homography, angle, focal);

	printTime("First bundle adjustment done in ");
}

// Second full bundle adjustment to minimize the reprojection errors
void ConcentricMosaic::performBundleAdj() {
	Vec3d axis(0, 1, 0);
	optimizeReprojection(graph, angle, axis, focal);
	graph.clear();

	// Scaling by the focal length
	double F[3][3] = {
		{ focal, 0, 0 },
		{ 0, focal, 0 },
		{ 0, 0, 1 }
	};
	Mat mFocal = Mat(3, 3, CV_64F, F);

	// Camera are centered at (0,0,0) in world's coordinate,
	// hence we only need a 3×3 projection matrix for each of them
	axis = Vec3d(0, 1, 0);
	for (int i = 0; i < n; ++i)
		projection[i] = mTranslate * mFocal * rotationFromAxis(axis, angle[i]);

	printTime("Second bundle adjustment done in ");
}

/*!********************************************************************************
 * STEP 3: Create the cylindrical map
 **********************************************************************************/

void ConcentricMosaic::createWorldMap(const vector<Mat> & img) {
	// Find the top and bottom points to display on the cylinder
	double yMin = 0, yMax = 0;
	for (int i = 0; i < n; ++i) {
		// Unproject the two points on the center-top and center-bottom of each image
		double w = img[i].cols, h = img[i].rows;
		double T[3][2] = { { w / 2, w / 2 }, { 0, h }, { 1, 1 } };

		// Corners (extremities) of the mapped image
		Mat corners = Mat(3, 2, CV_64F, T);
		corners = (projection[i].inv() * corners).t();

		// Extract minimum and maximum height.
		for (int i = 0; i < corners.rows; ++i) {
			double *p = corners.ptr<double> (i);
			double v = p[1] / sqrt(p[0] * p[0] + p[2] * p[2]);
			yMin = min(yMin, v);
			yMax = max(yMax, v);
		}
	}

	if (not quiet)
		cout << "Cylinder's extrema: [" << yMin << ", " << yMax << "]" << endl;

	printTime("Projection extrema found in ");

	// Create a matrix such that mWorld[x,y] = (x',y',z') the 3D point to project to each camera
	double ry = (yMax - yMin) / finalHeight;
	finalWidth = round(finalHeight * 2 * Pi / (yMax - yMin));

	// The shape is suited for future use of efficient matrix multiplications
	mWorld = Mat(3, finalHeight * finalWidth, CV_64F);
	double *u = mWorld.ptr<double> (0);
	double *v = mWorld.ptr<double> (1);
	double *w = mWorld.ptr<double> (2);

	for (int y = 0, i = 0; y < finalHeight; ++y)
		for (int x = 0; x < finalWidth; ++x, ++i) {
			u[i] = sin(toAngle(x));
			v[i] = y * ry + yMin;
			w[i] = cos(toAngle(x));
		}

	printTime("World map created in ");
}

/*!********************************************************************************
 * STEP 4: Warp the camera images with either a 2-band blending scheme or no blending
 **********************************************************************************/

void ConcentricMosaic::warp(const Mat & img, int i, Mat & warped) {
	//////////////////////////////////////////////////
	// CYLINDRICAL MAPPING
	//////////////////////////////////////////////////
	Mat map[2] = {
		Mat(1, finalWidth * finalHeight, CV_64F, Scalar(-1)),
		Mat(1, finalWidth * finalHeight, CV_64F, Scalar(-1)) };

	// Apply the projection, and then divide by the third coordinate.
	// But it would produce two pictures after remaping, so we filter the reversed one
	Mat points = projection[i] * mWorld;
	Mat mask = points.row(2) > 0;
	for (int j = 0; j < 2; ++j) {
		Mat view = points.row(j);
		divide(view, points.row(2), view);
		view.copyTo(map[j], mask);
		map[j].reshape(1, finalHeight).convertTo(map[j], CV_32F);
	}

	// Yield and image such that warped[x,y] = img[map[0][x,y], map[1][x,y]]
	remap(img, warped, map[0], map[1], INTER_CUBIC, BORDER_TRANSPARENT);
}

inline double modulo(double x, double y) {
	return x < 0 ? fmod(x, y) + y : fmod(x, y);
}

void ConcentricMosaic::borders(int i, double phi, vector<Vec3i> & mask) {
	// Here we assume a Right → Left movement, i.e. that we should have theta_2 < theta_1 < theta_0
	double theta0 = modulo(- angle[(i + n - 1) % n] + phi, 2 * Pi);
	double theta1 = modulo(- angle[i] + phi, 2 * Pi);
	double theta2 = modulo(- angle[(i + 1) % n] + phi, 2 * Pi);

	if (theta2 > theta1) {
		mask.push_back(Vec3i(toColumn(theta2 - 2 * Pi), toColumn(theta1), toColumn(theta0)));
		mask.push_back(Vec3i(toColumn(theta2), toColumn(theta1 + 2 * Pi), toColumn(theta0 + 2 * Pi)));
	} else if (theta1 > theta0) {
		mask.push_back(Vec3i(toColumn(theta2), toColumn(theta1), toColumn(theta0 + 2 * Pi)));
		mask.push_back(Vec3i(toColumn(theta2 - 2 * Pi), toColumn(theta1 - 2 * Pi), toColumn(theta0)));
	} else
		mask.push_back(Vec3i(toColumn(theta2), toColumn(theta1), toColumn(theta0)));
}

/*!********************************************************************************
 * Assemble the images with or withour blending
 **********************************************************************************/

void ConcentricMosaic::blend(const Mat & img, int i, double phi, Mat & dst) {
	// Warp the image onto a cylinder, using float values
	Mat tmp, warped;
	img.convertTo(tmp, CV_32FC3);
	warp(tmp, i, warped);
	tmp.release();

	// Retrieve the weight borders (weighting is done in the cylindrical space directly
	vector<Vec3i> mask;
	borders(i, phi, mask);

	//////////////////////////////////////////////////
	// LOW PASS
	//////////////////////////////////////////////////

	Mat lowPass;
	double sigma = 2; // Gaussian kernel stddev
	GaussianBlur(warped, lowPass, Size(0, 0), sigma, sigma, BORDER_CONSTANT);

	// Weighted contribution, added to the final image
	for (int j = 0; j < mask.size(); ++j) {
		for (int x = max(0, mask[j][0]); x < mask[j][1] && x < lowPass.cols; ++x) {
			double w = (double) (x - mask[j][0]) / (mask[j][1] - mask[j][0]);
			dst.col(x) += lowPass.col(x) * w;
		}
		for (int x = max(0, mask[j][1]); x < mask[j][2] && x < lowPass.cols; ++x) {
			double w = (double) (mask[j][2] - x) / (mask[j][2] - mask[j][1]);
			dst.col(x) += lowPass.col(x) * w;
		}
	}

	//////////////////////////////////////////////////
	// HIGH PASS
	//////////////////////////////////////////////////

	// Add the high frequencies from the image with the maximum weigh
	warped = warped - lowPass;
	for (int j = 0; j < mask.size(); ++j) {
		int x0 = max((int) round((mask[j][0] + mask[j][1]) / 2.), 0);
		int x1 = min((int) round((mask[j][1] + mask[j][2]) / 2.), finalWidth);
		if (x0 < finalWidth && x1 >= 0)
			dst.colRange(x0, x1) += warped.colRange(x0, x1);
	}
}

void ConcentricMosaic::assemble(const Mat & img, int i, double phi, Mat & dst) {
	// Warp the image onto a cylinder
	Mat warped;
	//warp(img, i, dst);
	//return;

	//////////////////////////////////////////////////
	// Alternative version : stitch over non-overlapping bands
	//////////////////////////////////////////////////
	warp(img, i, warped);

	// Retrieve the weight borders (weighting is done in the cylindrical space directly
	vector<Vec3i> mask;
	borders(i, phi, mask);

	// Add the information from the image with the maximum weigh only
	for (int j = 0; j < mask.size(); ++j) {
		int x0 = max((int) round((mask[j][0] + mask[j][1]) / 2.), 0);
		int x1 = min((int) round((mask[j][1] + mask[j][2]) / 2.), warped.cols);
		if (x0 < warped.cols && x1 >= 0)
			dst.colRange(x0, x1) += warped.colRange(x0, x1);
	}
}
