#ifndef CONCENTRICMOSAIC_H_
#define CONCENTRICMOSAIC_H_

// Include
#include <opencv2/core/core.hpp>
#include <opencv2/features2d/features2d.hpp>
#include "definition.h"
using namespace cv;

// Typedef
typedef vector<pair<Mat, Mat> > MatchGraph;

// Class definition
class ConcentricMosaic
{
public:
	ConcentricMosaic(double ratio = 0.6, double distance = 0.1, bool quiet = false);

	// Main function that stitches the images together
	void compute(const vector<Mat> & center, const vector<Mat> & offset, double radius,
		int height, bool blending = true, const string & output = "result.png");

private:
	// STEP 1
	void findKeypoints(const vector<Mat> & img);
	void findMatches();

	// STEP 2
	void initalizeBundleAdj();
	void performBundleAdj();

	// STEP 3
	void createWorldMap(const vector<Mat> & img);

	// STEP 4
	void warp(const Mat & img, int i, Mat & dst);
	void borders(int i, double phi, vector<Vec3i> & mask);

	void blend(const Mat & img, int i, double phi, Mat & dst);
	void assemble(const Mat & img, int i, double phi, Mat & dst);

	// Miscellaneous
	double toAngle(int x);
	int toColumn(double theta);

	void printTime(const string & message);

private:
	typedef void (ConcentricMosaic::*StitchFunction)(const Mat & img, int i, double phi, Mat & dst);

	// SIFT features
	vector<vector<KeyPoint> > keypoints;
	vector<Mat> descriptors;

	// Bundle adjustment parameters
	MatchGraph graph;
	vector<double> angle;
	double focal;

	vector<Mat> projection;

	// Matches filtering parameters
	double matchRatioThreshold;
	double matchDistanceThreshold;
	// Number of images in the input (img_0, ... img_{n-1}, img_n = img_0)
	int n;
	// Mosaic dimensions
	int finalWidth;
	int finalHeight;
	// Camera snapshots dimensions
	int originalWidth;
	int originalHeight;
	// Program arguments
	const bool quiet;
	double time;

	// Translation matrix to center the images
	Mat mTranslate;
	// World points coordinates for the cylindrical mapping
	Mat mWorld;
};

// Inline methods
inline double ConcentricMosaic::toAngle(int x) {
	return 2 * Pi * x / (finalWidth - 1);
}

inline int ConcentricMosaic::toColumn(double theta) {
	//theta = (theta < 0 ? fmod(theta, 2 * Pi) + 2 * Pi : fmod(theta, 2 * Pi));
	return round((finalWidth - 1) * theta / (2 * Pi));
}

#endif /*CONCENTRICMOSAIC_H_*/
